# ----------------------------------------------------------------------
#    Copyright (C) 2013 Kshitij Gupta <kgupta8592@gmail.com>
#    Copyright (C) 2014-2024 Christian Boltz <apparmor@cboltz.de>
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------
import apparmor.aa as apparmor


class Prof:
    def __init__(self, filename):
        apparmor.init_aa()
        self.active_profiles = apparmor.active_profiles
        self.include = apparmor.include
        self.filename = filename


class CleanProf:
    def __init__(self, same_file, profile, other):
        # If same_file we're basically comparing the file against itself to check superfluous rules
        self.same_file = same_file
        self.profile = profile
        self.other = other

    def compare_profiles(self):
        deleted = 0

        deleted += self.other.active_profiles.delete_preamble_duplicates(self.other.filename)

        for profile in self.profile.active_profiles.get_all_profiles():
            deleted += self.remove_duplicate_rules(profile)

        return deleted

    def remove_duplicate_rules(self, program):
        # Process the profile of the program

        deleted = 0

        # remove duplicate rules from the preamble
        deleted += self.profile.active_profiles.delete_preamble_duplicates(self.profile.filename)

        # Process every hat in the profile individually
        for full_profile in sorted(self.profile.active_profiles.get_profile_and_childs(program)):
            includes = self.profile.active_profiles[full_profile]['inc_ie'].get_all_full_paths(apparmor.profile_dir)

            # Clean up superfluous rules from includes in the other profile
            for inc in includes:
                if not self.profile.include.get(inc, {}).get(inc, False):
                    apparmor.load_include(inc)
                if self.other.active_profiles.profile_exists(full_profile):
                    deleted += apparmor.delete_all_duplicates(self.other.active_profiles[full_profile], inc, apparmor.ruletypes)

            # Clean duplicate rules in other profile
            for ruletype in apparmor.ruletypes:
                if not self.same_file:
                    if self.other.active_profiles.profile_exists(full_profile):
                        deleted += self.other.active_profiles[full_profile][ruletype].delete_duplicates(self.profile.active_profiles[full_profile][ruletype])
                else:
                    deleted += self.other.active_profiles[full_profile][ruletype].delete_duplicates(None)

        return deleted



# ----------------------------------------------------------------------
#    Copyright (C) 2024 Christian Boltz
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------

import re

from apparmor.common import AppArmorBug, AppArmorException

from apparmor.aare import AARE
from apparmor.regex import RE_PROFILE_PIVOT_ROOT, RE_PROFILE_NAME, RE_PROFILE_PATH_OR_VAR, strip_quotes
from apparmor.rule import BaseRule, BaseRuleset, parse_modifiers, logprof_value_or_all, quote_if_needed

from apparmor.translations import init_translation

_ = init_translation()

RE_PIVOT_ROOT_DETAILS = re.compile(
    r'^\s*'
    + r'(\s+oldroot=' + RE_PROFILE_PATH_OR_VAR % 'oldroot'  + r')?'  # noqa: E221
    + r'(\s+'         + RE_PROFILE_PATH_OR_VAR % 'newroot'  + r')?'  # noqa: E221
    + r'(\s+->\s+'    + RE_PROFILE_NAME % 'profile_name'    + r')?'  # noqa: E221
    + r'\s*$'
)


class PivotRootRule(BaseRule):
    '''Class to handle and store a single pivot_root rule'''

    # Nothing external should reference this class, all external users
    # should reference the class field PivotRootRule.ALL
    class __PivotRootAll(object):
        pass

    ALL = __PivotRootAll

    rule_name = 'pivot_root'
    _match_re = RE_PROFILE_PIVOT_ROOT

#            PIVOT ROOT RULE = [ QUALIFIERS ] pivot_root [ oldroot=OLD PUT FILEGLOB ] [ NEW ROOT FILEGLOB ] [ ’->’ PROFILE NAME ]
    def __init__(self, oldroot, newroot, profile_name, audit=False, deny=False, allow_keyword=False,
                 comment='', log_event=None, priority=None):

        super().__init__(audit=audit, deny=deny,
                         allow_keyword=allow_keyword,
                         comment=comment,
                         log_event=log_event, priority=priority)

        self.oldroot,      self.all_oldroots        = self._aare_or_all(oldroot,        'oldroot',        True,  log_event)  # noqa: E221
        self.newroot,      self.all_newroots        = self._aare_or_all(newroot,        'newroot',        True,  log_event)  # noqa: E221
        self.profile_name, self.all_profile_names   = self._aare_or_all(profile_name,   'profile_name',   False, log_event)  # noqa: E221

        self.can_glob = not self.all_newroots
        self.can_glob_ext = False
        self.can_edit = not self.all_newroots

    @classmethod
    def _create_instance(cls, raw_rule, matches):
        '''parse raw_rule and return instance of this class'''

        priority, audit, deny, allow_keyword, comment = parse_modifiers(matches)

        rule_details = ''
        if matches.group('details'):
            rule_details = matches.group('details')

            parsed = RE_PIVOT_ROOT_DETAILS.search(rule_details)

            if not parsed:
                raise AppArmorException('Cannot parse pivot_root rule ' + raw_rule)

            r = parsed.groupdict()

            if r['oldroot']:
                oldroot = strip_quotes(r['oldroot'])
            else:
                oldroot = cls.ALL

            if r['newroot']:
                newroot = strip_quotes(r['newroot'])
            else:
                newroot = cls.ALL

            if r['profile_name']:
                profile_name = strip_quotes(r['profile_name'])
            else:
                profile_name = cls.ALL

        else:
            oldroot = cls.ALL
            newroot = cls.ALL
            profile_name = cls.ALL

        return cls(oldroot=oldroot, newroot=newroot, profile_name=profile_name,
                   audit=audit, deny=deny, allow_keyword=allow_keyword, comment=comment, priority=priority)

    def get_clean(self, depth=0):
        space = '  ' * depth

        if self.all_oldroots:
            oldroot = ''
        elif self.oldroot:
            oldroot = ' oldroot=' + quote_if_needed(self.oldroot.regex)
        else:
            raise AppArmorBug('Empty oldroot in pivot_root rule')

        if self.all_newroots:
            newroot = ''
        elif self.newroot:
            newroot = ' ' + quote_if_needed(self.newroot.regex)
        else:
            raise AppArmorBug('Empty newroot in pivot_root rule')

        if self.all_profile_names:
            profile_name = ''
        elif self.profile_name:
            profile_name = ' -> ' + quote_if_needed(self.profile_name.regex)
        else:
            raise AppArmorBug('Empty profile_name in pivot_root rule')

        return f'{space}{self.modifiers_str()}pivot_root{oldroot}{newroot}{profile_name},{self.comment}'

    def _is_covered_localvars(self, other_rule):
        if not self._is_covered_aare(self.oldroot, self.all_oldroots, other_rule.oldroot, other_rule.all_oldroots, 'oldroot'):
            return False

        if not self._is_covered_aare(self.newroot, self.all_newroots, other_rule.newroot, other_rule.all_newroots, 'newroot'):
            return False

        if not self._is_covered_aare(self.profile_name, self.all_profile_names, other_rule.profile_name, other_rule.all_profile_names, 'profile_name'):
            return False

        # still here? -> then it is covered
        return True

    def _is_equal_localvars(self, other_rule, strict):
        if not self._is_equal_aare(self.oldroot, self.all_oldroots, other_rule.oldroot, other_rule.all_oldroots, 'oldroot'):
            return False

        if not self._is_equal_aare(self.newroot, self.all_newroots, other_rule.newroot, other_rule.all_newroots, 'newroot'):
            return False

        if not self._is_equal_aare(self.profile_name, self.all_profile_names, other_rule.profile_name, other_rule.all_profile_names, 'profile_name'):
            return False

        return True

    def glob(self):
        '''Change newroot path to next possible glob'''
        if self.all_newroots:
            return

        self.newroot = self.newroot.glob_path()
        self.raw_rule = None

    def edit_header(self):
        if self.all_newroots:
            raise AppArmorBug('Attemp to edit pivot_root rule without newroot limitations')

        return (_('Enter new newroot: '), self.newroot.regex)

    def validate_edit(self, newpath):
        if self.all_newroots:
            raise AppArmorBug('Attemp to edit pivot_root rule without newroot limitations')

        newpath = AARE(newpath, True)  # might raise AppArmorException if the new path doesn't start with / or a variable
        return newpath.match(self.newroot)

    def store_edit(self, newpath):
        if self.all_newroots:
            raise AppArmorBug('Attemp to edit pivot_root rule without newroot limitations')

        self.newroot = AARE(newpath, True)  # might raise AppArmorException if the new path doesn't start with / or a variable
        self.raw_rule = None

    def _logprof_header_localvars(self):

        oldroot = logprof_value_or_all(self.oldroot, self.all_oldroots)
        newroot = logprof_value_or_all(self.newroot, self.all_newroots)
        profile_name = logprof_value_or_all(self.profile_name, self.all_profile_names)

        return (
            _('Old root'), oldroot,
            _('New root'), newroot,
            _('Target profile'), profile_name,
        )

    @staticmethod
    def hashlog_from_event(hl, ev):
        # TODO: can the log contain the target profile?
        hl[ev['src_name']][ev['name']] = True

    @classmethod
    def from_hashlog(cls, hl):
        for oldroot, newroot in BaseRule.generate_rules_from_hashlog(hl, 2):
            yield cls(oldroot, newroot, cls.ALL, log_event=True)


class PivotRootRuleset(BaseRuleset):
    '''Class to handle and store a collection of pivot_root rules'''



# ----------------------------------------------------------------------
#    Copyright (C) 2013 Kshitij Gupta <kgupta8592@gmail.com>
#    Copyright (C) 2014 Christian Boltz <apparmor@cboltz.de>
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------

import re

from apparmor.common import AppArmorBug, AppArmorException
from apparmor.regex import RE_PROFILE_CAP
from apparmor.rule import BaseRule, BaseRuleset, logprof_value_or_all, parse_modifiers
from apparmor.translations import init_translation

_ = init_translation()

capability_keywords = [
    'audit_control', 'audit_read', 'audit_write', 'block_suspend', 'bpf', 'checkpoint_restore',
    'chown', 'dac_override', 'dac_read_search', 'fowner', 'fsetid', 'ipc_lock', 'ipc_owner',
    'kill', 'lease', 'linux_immutable', 'mac_admin', 'mac_override', 'mknod', 'net_admin',
    'net_bind_service', 'net_broadcast', 'net_raw', 'perfmon', 'setfcap', 'setgid', 'setpcap',
    'setuid', 'syslog', 'sys_admin', 'sys_boot', 'sys_chroot', 'sys_module', 'sys_nice',
    'sys_pacct', 'sys_ptrace', 'sys_rawio', 'sys_resource', 'sys_time', 'sys_tty_config',
    'wake_alarm']


class CapabilityRule(BaseRule):
    """Class to handle and store a single capability rule"""

    # Nothing external should reference this class, all external users
    # should reference the class field CapabilityRule.ALL
    class __CapabilityAll:
        pass

    ALL = __CapabilityAll

    rule_name = 'capability'
    _match_re = RE_PROFILE_CAP

    def __init__(self, cap_list, audit=False, deny=False, allow_keyword=False,
                 comment='', log_event=None, priority=None):

        super().__init__(audit=audit, deny=deny, allow_keyword=allow_keyword,
                         comment=comment, log_event=log_event, priority=priority)
        # Because we support having multiple caps in one rule,
        # initializer needs to accept a list of caps.
        self.all_caps = False
        if cap_list == self.ALL:
            self.all_caps = True
            self.capability = set()
        else:
            if isinstance(cap_list, str):
                cap_list = [cap_list]

            if isinstance(cap_list, list):
                if not cap_list:
                    raise AppArmorBug('Passed empty capability list to %s: %s' % (type(self).__name__, str(cap_list)))
                for cap in cap_list:
                    if not cap.strip():
                        # make sure none of the cap_list arguments are blank, in
                        # case we decide to return one cap per output line
                        raise AppArmorBug('Passed empty capability to %s: %s' % (type(self).__name__, str(cap_list)))
                    if cap not in capability_keywords:
                        raise AppArmorException('Passed unknown capability to %s: %s' % (type(self).__name__, cap))
                self.capability = set(cap_list)
            else:
                raise AppArmorBug('Passed unknown object to %s: %s' % (type(self).__name__, str(cap_list)))

    @classmethod
    def _create_instance(cls, raw_rule, matches):
        """parse raw_rule and return instance of this class"""

        priority, audit, deny, allow_keyword, comment = parse_modifiers(matches)

        if matches.group('capability'):
            capability = matches.group('capability').strip()
            capability = re.split("[ \t]+", capability)
        else:
            capability = cls.ALL

        return cls(capability, audit=audit, deny=deny,
                   allow_keyword=allow_keyword,
                   comment=comment, priority=priority)

    def get_clean(self, depth=0):
        """return rule (in clean/default formatting)"""

        space = '  ' * depth
        if self.all_caps:
            return ('%s%scapability,%s' % (space, self.modifiers_str(), self.comment))
        else:
            caps = ' '.join(self.capability).strip()  # XXX return multiple lines, one for each capability, instead?
            if caps:
                return ('%s%scapability %s,%s' % (space, self.modifiers_str(), ' '.join(sorted(self.capability)), self.comment))
            else:
                raise AppArmorBug("Empty capability rule")

    def _is_covered_localvars(self, other_rule):
        """check if other_rule is covered by this rule object"""

        if not self._is_covered_list(self.capability, self.all_caps, other_rule.capability, other_rule.all_caps, 'capability'):
            return False

        # still here? -> then it is covered
        return True

    def _is_equal_localvars(self, other_rule, strict):
        """compare if rule-specific variables are equal"""

        if (self.capability != other_rule.capability
                or self.all_caps != other_rule.all_caps):
            return False

        return True

    def severity(self, sev_db):
        if self.all_caps:
            severity = sev_db.rank_capability('__ALL__')
        else:
            severity = -1
            for cap in self.capability:
                sev = sev_db.rank_capability(cap)
                if isinstance(sev, int):  # type check avoids breakage caused by 'unknown'
                    severity = max(severity, sev)

        if severity == -1:
            severity = sev  # effectively 'unknown'

        return severity

    def _logprof_header_localvars(self):
        cap_txt = logprof_value_or_all(self.capability, self.all_caps)

        return _('Capability'), cap_txt

    @staticmethod
    def hashlog_from_event(hl, ev):
        hl[ev['name']] = True

    @classmethod
    def from_hashlog(cls, hl):
        for cap in hl.keys():
            yield cls(cap, log_event=True)


class CapabilityRuleset(BaseRuleset):
    """Class to handle and store a collection of capability rules"""

    def get_glob(self, path_or_rule):
        """Return the next possible glob. For capability rules, that's always "capability," (all capabilities)"""
        return 'capability,'



# ----------------------------------------------------------------------
#    Copyright (C) 2020 Christian Boltz <apparmor@cboltz.de>
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------

from apparmor.common import AppArmorBug
from apparmor.regex import RE_ABI
from apparmor.rule.include import IncludeRule, IncludeRuleset
from apparmor.translations import init_translation

_ = init_translation()


# abi and include rules have a very similar syntax
# base AbiRule on IncludeRule to inherit most of its behaviour
class AbiRule(IncludeRule):
    """Class to handle and store a single abi rule"""

    rule_name = 'abi'
    _match_re = RE_ABI

    def __init__(self, path, ifexists, ismagic, audit=False, deny=False, allow_keyword=False,
                 comment='', log_event=None, priority=None):

        super().__init__(path, ifexists, ismagic,
                         audit=audit, deny=deny, allow_keyword=allow_keyword,
                         comment=comment, log_event=log_event, priority=priority)

        # abi doesn't support 'if exists'
        if ifexists:
            raise AppArmorBug('Attempt to use %s rule with if exists flag' % self.__class__.__name__)

    def get_clean(self, depth=0):
        """return rule (in clean/default formatting)"""

        space = '  ' * depth

        if self.ismagic:
            return ('%s%s <%s>,%s' % (space, self.rule_name, self.path, self.comment))
        else:
            return ('%s%s "%s",%s' % (space, self.rule_name, self.path, self.comment))

    def _logprof_header_localvars(self):
        return _('Abi'), self.get_clean()


class AbiRuleset(IncludeRuleset):
    """Class to handle and store a collection of abi rules"""



#
# ----------------------------------------------------------------------
#    Copyright (C) 2025 Canonical, Ltd.
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------

import ast
import re

from apparmor.common import AppArmorException, AppArmorBug
from apparmor.regex import strip_quotes, RE_PROFILE_CONDITIONAL_START, RE_PROFILE_CONDITIONAL_ELSE, RE_ALL_VARIABLES, RE_BOOLEAN_OP, RE_ID_OR_VAR, RE_CONDITION, RE_COMPARE_OP_QUOTED
from apparmor.rule import BaseRule, BaseRuleset, parse_comment, quote_if_needed
from apparmor.translations import init_translation

_ = init_translation()


class ConditionalBlock(BaseRule):
    """Class to handle and store a conditional block containing if,
    and optionally, else ifs and else"""

    _match_re = None
    result = False

    def __init__(self, raw_rule, prof_storage, audit=False, deny=False, allow_keyword=False,
                 comment='', log_event=None, priority=None):

        super().__init__(audit=audit, deny=deny, allow_keyword=allow_keyword,
                         comment=comment, log_event=log_event, priority=priority)

        # conditional blocks don't support priority, allow keyword, audit or deny - yet
        self.ensure_modifiers_not_supported()

        self.cond_list = []
        conditional = ConditionalStart.create_instance(raw_rule)
        self.result = conditional.evaluate(prof_storage)
        self.cond_list.append(conditional)

    def store_profile_data(self, profile_data):
        self.cond_list[-1].profile_data = profile_data

    def add_conditional(self, raw_rule, prof_storage):
        conditional = ConditionalElse.create_instance(raw_rule)
        result = conditional.evaluate(prof_storage)
        self.cond_list.append(conditional)
        if self.result:
            self.result = False
        else:
            self.result = result

    def get_clean(self, depth=0):
        clean = ''
        for cond in self.cond_list:
            if clean:
                clean += ' '
            clean += cond.get_clean(depth)
        return clean

    @classmethod
    def _create_instance(cls, raw_rule, matches):
        raise NotImplementedError("'%s' is not supposed to be called directly" % (str(cls)))

    def _is_covered_localvars(self, other_rule):
        """check if other_rule is covered by this rule object"""
        if len(self.cond_list) != len(other_rule.cond_list):
            return False
        else:
            for idx in range(len(self.cond_list)):
                if not self.cond_list[idx].is_covered(other_rule.cond_list[idx]):
                    return False
        return True

    def _is_equal_localvars(self, other_rule, strict):
        """compare if rule-specific conditionals are equal"""
        if len(self.cond_list) != len(other_rule.cond_list):
            return False
        else:
            for idx in range(len(self.cond_list)):
                if not self.cond_list[idx].is_equal(other_rule.cond_list[idx]):
                    return False
        return True

    def _logprof_header_localvars(self):
        conditions = []
        for cond in self.cond_list:
            conditions.append(cond.logprof_header())
        return _('ConditionalBlock'), (conditions)


class ConditionalRule(BaseRule):
    """Class to handle and store a single conditional rule"""

    IF = 1
    ELSEIF = 2
    ELSE = 3
    rule_name = 'conditional'
    _match_re = re.compile(RE_CONDITION)

    def __init__(self, condition, ast_tree,
                 audit=False, deny=False, allow_keyword=False,
                 comment='', log_event=None, priority=None):

        super().__init__(audit=audit, deny=deny, allow_keyword=allow_keyword,
                         comment=comment, log_event=log_event, priority=priority)

        # conditionals don't support priority, allow keyword, audit or deny
        self.ensure_modifiers_not_supported()

        if condition not in [self.IF, self.ELSEIF, self.ELSE]:
            raise AppArmorBug('Passed invalid condition to %s: %s' % (self.__class__.__name__, condition))
        if ast_tree is not None and not isinstance(ast_tree, AppArmorAst):
            raise AppArmorBug('Passed invalid AST tree type to %s: %s' % (self.__class__.__name__, type(ast_tree)))

        self.condition = condition
        self.ast_tree = ast_tree

    @classmethod
    def _create_instance(cls, raw_rule, matches):
        """parse raw_rule and return instance of this class"""

        if cls == ConditionalStart:
            conditional = ConditionalRule.IF
        else:
            if matches.group('if'):
                conditional = ConditionalRule.ELSEIF
            else:
                conditional = ConditionalRule.ELSE

        comment = parse_comment(matches)

        expr = cls._match_re.search(raw_rule)
        ast_tree = None
        if conditional != ConditionalRule.ELSE:
            ast_tree = AppArmorAst(expr.group('expr'))

        return cls(conditional, ast_tree, comment)

    def evaluate(self, prof_storage):
        # else should always evaluate to true, since that's the
        # default if all previous "ifs" evaluated to false
        if self.condition == ConditionalRule.ELSE:
            return True

        return self.ast_tree.evaluate(prof_storage)

    def get_clean(self, depth=0):
        """return rule (in clean/default formatting)"""

        space = '  ' * depth
        leading_space = ''
        if self.condition == ConditionalRule.IF:
            conditional = 'if '
            leading_space = space
        elif self.condition == ConditionalRule.ELSEIF:
            conditional = 'else if '
        elif self.condition == ConditionalRule.ELSE:
            conditional = 'else '
        else:
            raise AppArmorBug('Invalid condition type in %s' % (self.__class__.__name__))

        expr = ''
        if self.ast_tree:
            expr = self.ast_tree.get_clean()

        data = []
        data.append('%s%s%s{' % (leading_space, conditional, expr))

        for profname in self.profile_data:
            if self.profile_data[profname]['in_cond']:
                from apparmor.aa import write_piece
                data.extend(write_piece(self.profile_data, depth + 1, profname, profname))
            else:
                data += self.profile_data[profname].get_rules_clean(depth + 1)

        data.append('%s}' % space)
        return '\n'.join(data)

    def _is_covered_localvars(self, other_rule):
        """check if other_rule is covered by this rule object"""
        # conditional is only covered if equal
        if self.is_equal(other_rule):
            return True
        return False

    def _is_equal_localvars(self, rule_obj, strict):
        """compare if rule-specific conditionals are equal"""
        if self.condition != rule_obj.condition:
            return False
        if self.profile_data != rule_obj.profile_data:
            return False
        return self.ast_tree.is_equal(rule_obj.ast_tree)

    def _logprof_header_localvars(self):
        return _('Conditional'), self.get_clean()


class ConditionalStart(ConditionalRule):
    """Class to handle and store a single conditional rule"""

    _match_re = RE_PROFILE_CONDITIONAL_START


class ConditionalElse(ConditionalRule):
    """Class to handle and store a single conditional rule"""

    _match_re = RE_PROFILE_CONDITIONAL_ELSE


class ConditionalBlockset(BaseRuleset):
    """Class to handle and store a collection of conditional rule blocks"""


class AppArmorAst():
    astcomp_to_string = {
        ast.Eq: '==',
        ast.NotEq: '!=',
        ast.Lt: '<',
        ast.LtE: '<=',
        ast.Gt: '>',
        ast.GtE: '>=',
        ast.In: 'in',
    }

    def __init__(self, expr):
        self.tree = ast.parse(self.transform_cond(expr))

    def get_clean(self):
        noop, expr = self.get_clean_tree(self.tree.body[0])
        expr += ' '
        return expr

    def get_clean_tree(self, node):
        if isinstance(node, ast.Expr):
            node = node.value
        if isinstance(node, ast.BoolOp):
            op = ''
            clean = ''
            if isinstance(node.op, ast.And):
                op = 'and'
            else:
                op = 'or'
            for value in node.values:
                if clean:
                    clean += ' ' + op + ' '
                ret_op, ret_clean = self.get_clean_tree(value)
                if ret_op == 'or' and op == 'and':
                    ret_clean = '(' + ret_clean + ')'
                clean += ret_clean
            return op, clean
        elif isinstance(node, ast.UnaryOp):
            if not isinstance(node.op, ast.Not):
                raise AppArmorBug('Invalid unary operation in %s' % (self.__class__.__name__))
            op, child = self.get_clean_tree(node.operand)
            if op == 'not':  # remove canceling nots
                return 'val', '%s' % (child[len('not '):])
            return 'not', 'not %s' % (child)
        elif isinstance(node, ast.Constant):
            val = quote_if_needed(strip_quotes(node.value))  # strip first because it can be quoted but not need it
            term = Term.create_instance(val)
            return 'val', str(term)
        elif isinstance(node, ast.Name):
            return 'func', node.id
        elif isinstance(node, ast.Call):
            noop, name = self.get_clean_tree(node.func)
            if name != 'defined':
                raise AppArmorBug('Invalid function name in %s' % (self.__class__.__name__))
            noop, var = self.get_clean_tree(node.args[0])
            return 'defined', '%s %s' % (name, var)
        elif isinstance(node, ast.Compare):
            noop, left = self.get_clean_tree(node.left)
            noop, right = self.get_clean_tree(node.comparators[0])
            op = self.astcomp_to_string[type(node.ops[0])]
            return 'cmp', '%s %s %s' % (left, op, right)
        else:
            raise AppArmorBug('Unsupported node type in %s' % (self.__class__.__name__))

    def evaluate(self, prof_storage):
        noop, result = self.evaluate_tree(self.tree.body[0], prof_storage, True)
        return result

    def evaluate_tree(self, node, prof_storage, resolve=False):
        result = None
        if isinstance(node, ast.Expr):
            node = node.value
        if isinstance(node, ast.BoolOp):
            op = node.op
            for value in node.values:
                ret_op, ret_result = self.evaluate_tree(value, prof_storage, True)
                if result is None:
                    result = ret_result
                else:
                    if isinstance(op, ast.And):
                        result = result and ret_result
                    else:
                        result = result or ret_result
            return op, result
        elif isinstance(node, ast.UnaryOp):
            if not isinstance(node.op, ast.Not):
                raise AppArmorBug('Invalid unary operation in %s' % (self.__class__.__name__))
            result = not self.evaluate_tree(node.operand, prof_storage, True)
            return node.op, result
        elif isinstance(node, ast.Constant):
            val = quote_if_needed(strip_quotes(node.value))  # strip first because it can be quoted but not need it
            term = Term.create_instance(val)
            if resolve:
                cond = BooleanCondition('', term)
                result = cond.evaluate(prof_storage)
            return term, result
        elif isinstance(node, ast.Name):
            return node.id, result
        elif isinstance(node, ast.Call):
            func, noop = self.evaluate_tree(node.func, prof_storage)
            if func != 'defined':
                raise AppArmorBug('Invalid function name in %s' % (self.__class__.__name__))
            # there should be only one arg
            variable, noop = self.evaluate_tree(node.args[0], prof_storage)
            cond = BooleanCondition(func, variable)
            return 'defined', cond.evaluate(prof_storage)
        elif isinstance(node, ast.Compare):
            left, noop = self.evaluate_tree(node.left, prof_storage)
            right, noop = self.evaluate_tree(node.comparators[0], prof_storage)
            op = self.astcomp_to_string[type(node.ops[0])]
            cond = CompareCondition(left, op, right)
            return 'cmp', cond.evaluate(prof_storage)
        else:
            raise AppArmorBug('Unsupported node type in %s' % (self.__class__.__name__))

    def compare_ast(self, node1, node2):
        if type(node1) is not type(node2):
            return False
        if isinstance(node1, ast.AST):
            for k, v in vars(node1).items():
                if k in ('lineno', 'col_offset', 'ctx', 'end_lineno', 'end_col_offset'):
                    continue
                if not self.compare_ast(v, getattr(node2, k)):
                    return False
            return True
        elif isinstance(node1, list):
            if len(node1) != len(node2):
                return False
            for i in range(len(node1)):
                if not self.compare_ast(node1[i], node2[i]):
                    return False
            return True
        else:
            return node1 == node2

    def is_equal(self, other_tree):
        return self.compare_ast(self.tree, other_tree.tree)

    def transform_cond(self, text):
        """Used to transform policy conditional into Python format, so ast can be used"""

        def boolean_op(match):
            not_op = match.group('boolean_not')
            defined = match.group('defined')
            var = match.group('var')
            var = '"%s"' % (var)
            if defined:
                var = 'defined(%s)' % (var)
            return '%s%s' % (not_op, var)

        def compare_op(match):
            not_op = match.group('compare_not')
            left = match.group('left')
            op = match.group('op')
            right = match.group('right')

            if match.group('id_left') and not (left.startswith('"') or left.endswith('"')):
                left = '"%s"' % (left)
            if match.group('id_right') and not (right.startswith('"') or right.endswith('"')):
                right = '"%s"' % (right)

            return '%s%s %s %s' % (not_op, left, op, right)

        replaced = re.sub(RE_BOOLEAN_OP % {'term': ''}, boolean_op, text)
        replaced = re.sub(RE_COMPARE_OP_QUOTED % {'term': ''}, compare_op, replaced)
        return replaced


class Term():
    match_re = re.compile(RE_ID_OR_VAR % {'label': ''})

    @classmethod
    def create_instance(cls, raw_term):
        """parse raw_term and return instance of this class"""
        matches = cls.match_re.search(raw_term)
        if not matches:
            raise AppArmorBug('Unable to parse term in %s' % (cls.__class__.__name__))
        if matches.group('id'):
            return Id(matches.group('id'))
        else:
            var_type = matches.group('var_type')
            varname = matches.group('varname')
            var = matches.group('var')
            return Variable(var_type, varname, var)


class Variable(Term):
    def __init__(self, var_type, varname, var):
        self.varname = varname
        self.var = var
        if var_type == '$':
            self.var_type = 'boolean'
        else:
            self.var_type = 'variable'

    def get_variable_rule(self, prof_storage):
        for rule in prof_storage[self.var_type].rules:
            filtered = RE_ALL_VARIABLES.search(rule.varname)
            if self.varname == filtered.group('varname'):
                return rule
        return None

    def get_set(self, prof_storage):
        variable_rule = self.get_variable_rule(prof_storage)
        if variable_rule is None:
            raise AppArmorException(_('Error retrieving variable %(var)s') % {'var': self.var})
        return variable_rule.values

    def __repr__(self):
        return self.var


class Id(Term):
    def __init__(self, value):
        self.value = value

    def get_set(self, prof_storage):
        return {self.value}

    def __repr__(self):
        return self.value


class BooleanCondition():
    def __init__(self, defined: str, variable: Term):
        if not isinstance(defined, str):
            raise AppArmorBug('Passed invalid defined value to %s: %s' % (self.__class__.__name__, defined))
        if not isinstance(variable, Term):
            raise AppArmorBug('Passed invalid variable type to %s: %s' % (self.__class__.__name__, type(variable)))

        self.defined = defined
        self.variable = variable

    def evaluate(self, prof_storage):
        matched = self.variable.get_variable_rule(prof_storage)

        if not self.defined:  # boolean op
            if self.variable.var_type == 'boolean':
                if matched:
                    return matched.value
                else:
                    raise AppArmorException(_('Cannot find previous declaration of %(var)s') % {'var': self.variable})
            else:
                raise AppArmorException(_('Unexpected variable in boolean operation: %(var)s') % {'var': self.variable})
        else:
            return bool(matched)


class CompareCondition():
    valid_ops = ['==', '!=', 'in', '>', '>=', '<', '<=']

    def __init__(self, left_term: Term, op: str, right_term: Term):
        if op not in self.valid_ops:
            raise AppArmorBug('Passed invalid op value to %s: %s' % (self.__class__.__name__, op))
        if not isinstance(left_term, Term):
            raise AppArmorBug('Passed invalid left term type to %s: %s' % (self.__class__.__name__, type(left_term)))
        if not isinstance(right_term, Term):
            raise AppArmorBug('Passed invalid right term type to %s: %s' % (self.__class__.__name__, type(right_term)))

        self.left_term = left_term
        self.op = op
        self.right_term = right_term

    def compare(self, op, lhs, rhs):
        if type(lhs) is not type(rhs):
            raise AppArmorBug('Trying to compare elements of different types in %s' % (self.__class__.__name__))

        if (op == '>'):
            return lhs > rhs
        elif (op == '>='):
            return lhs >= rhs
        elif (op == '<'):
            return lhs < rhs
        elif (op == '<='):
            return lhs <= rhs
        else:
            raise AppArmorBug('Invalid op in %s: %s' % (self.__class__.__name__, op))

    def evaluate(self, prof_storage):
        lhs = self.left_term.get_set(prof_storage)
        rhs = self.right_term.get_set(prof_storage)

        if not isinstance(lhs, set) or not isinstance(rhs, set):
            raise AppArmorBug('Passed invalid type for condition term in %s' % (self.__class__.__name__))

        converted_lhs = None
        converted_rhs = None

        if self.op == 'in':
            return lhs.issubset(rhs)
        elif self.op == '==':
            return lhs == rhs
        elif self.op == '!=':
            return lhs != rhs

        try:
            if len(lhs) == 1:
                converted_lhs = int(next(iter(lhs)))
        except ValueError:
            pass

        try:
            if len(rhs) == 1:
                converted_rhs = int(next(iter(rhs)))
        except ValueError:
            pass

        if converted_lhs is None and converted_rhs is None:  # sets
            return self.compare(self.op, lhs, rhs)
        elif converted_lhs is not None and converted_rhs is not None:  # numbers
            return self.compare(self.op, converted_lhs, converted_rhs)
        else:
            raise AppArmorException(_('Can only compare numbers with numbers'))



# ----------------------------------------------------------------------
#    Copyright (C) 2022 Canonical, Ltd.
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------

import re

from apparmor.regex import RE_PROFILE_MQUEUE, RE_PROFILE_NAME
from apparmor.common import AppArmorBug, AppArmorException
from apparmor.rule import BaseRule, BaseRuleset, check_and_split_list, logprof_value_or_all, parse_modifiers, quote_if_needed

# setup module translations
from apparmor.translations import init_translation
_ = init_translation()


access_keywords_read = ['r', 'read']
access_keywords_write = ['w', 'write']
access_keywords_rw = ['rw', 'wr']
access_keywords_other = ['create', 'open', 'delete', 'getattr', 'setattr']
access_keywords = access_keywords_read + access_keywords_write + access_keywords_rw + access_keywords_other

joint_access_keyword = r'\s*(' + '|'.join(access_keywords) + r')\s*'
RE_ACCESS_KEYWORDS = (joint_access_keyword  # one of the access_keyword or
                      + '|'                                            # or
                      + r'\(' + joint_access_keyword + '(' + r'(\s|,)+' + joint_access_keyword + ')*' + r'\)'  # one or more access_keyword in (...)
                      )

RE_MQUEUE_NAME = r'(?P<%s>(/\S*|\d*))'  # / + string for posix, or digits for sys
RE_MQUEUE_TYPE = r'(?P<%s>(sysv|posix))'  # type can be sysv or posix

RE_MQUEUE_DETAILS = re.compile(
    '^'
    + r'(\s+(?P<access>' + RE_ACCESS_KEYWORDS + '))?'  # optional access keyword(s)
    + r'(\s+(type=' + RE_MQUEUE_TYPE % 'mqueue_type' + '))?'  # optional type
    + r'(\s+(label=' + RE_PROFILE_NAME % 'label' + '))?'  # optional label
    + r'(\s+(' + RE_MQUEUE_NAME % 'mqueue_name' + '))?'  # optional mqueue name
    + r'\s*$')


class MessageQueueRule(BaseRule):
    '''Class to handle and store a single mqueue rule'''

    # Nothing external should reference this class, all external users
    # should reference the class field MessageQueueRule.ALL
    class __MessageQueueAll(object):
        pass

    ALL = __MessageQueueAll

    rule_name = 'mqueue'
    _match_re = RE_PROFILE_MQUEUE

    def __init__(self, access, mqueue_type, label, mqueue_name,
                 audit=False, deny=False, allow_keyword=False,
                 comment='', log_event=None, priority=None):

        super().__init__(audit=audit, deny=deny,
                         allow_keyword=allow_keyword,
                         comment=comment,
                         log_event=log_event,
                         priority=priority)

        self.access, self.all_access, unknown_items = check_and_split_list(access, access_keywords, self.ALL, type(self).__name__, 'access')
        if unknown_items:
            raise AppArmorException(_('Passed unknown access keyword to %s: %s') % (type(self).__name__, ' '.join(unknown_items)))

        self.label, self.all_labels = self._aare_or_all(label, 'label', is_path=False, log_event=log_event)
        self.mqueue_type, self.all_mqueue_types = self._aare_or_all(mqueue_type, 'type', is_path=False, log_event=log_event)
        if mqueue_name is not None:
            self.mqueue_name, self.all_mqueue_names = self._aare_or_all(mqueue_name, 'mqueue_name', is_path=False, log_event=log_event)
        else:
            self.mqueue_name, self.all_mqueue_names = None, self.ALL
        self.validate_mqueue_name()

    def validate_mqueue_name(self):
        # The regex checks if it starts with / or if it's numeric
        if self.all_mqueue_types or self.all_mqueue_names:
            return

        if self.mqueue_type.regex == 'sysv' and not self.mqueue_name.regex.isnumeric():
            raise AppArmorException(_('Queue name for SYSV must be a positive integer'))
        elif self.mqueue_type.regex == 'posix' and not self.mqueue_name.regex.startswith('/'):
            raise AppArmorException(_('Queue name for POSIX must begin with /'))

    @classmethod
    def _create_instance(cls, raw_rule, matches):
        '''parse raw_rule and return instance of this class'''

        priority, audit, deny, allow_keyword, comment = parse_modifiers(matches)

        rule_details = ''
        if matches.group('details'):
            rule_details = matches.group('details')

        if rule_details:
            details = RE_MQUEUE_DETAILS.search(rule_details)
            if not details:
                raise AppArmorException(_("Invalid or unknown keywords in 'mqueue %s" % rule_details))

            if details.group('access'):
                access = details.group('access')
                if access.startswith('(') and access.endswith(')'):
                    access = access[1:-1]
                access = access.replace(',', ' ').split()  # split by ',' or whitespace
            else:
                access = cls.ALL

            if details.group('mqueue_type'):
                mqueue_type = details.group('mqueue_type')
            else:
                mqueue_type = cls.ALL

            if details.group('label'):
                label = details.group('label')
            else:
                label = cls.ALL

            if details.group('mqueue_name'):
                mqueue_name = details.group('mqueue_name')
            else:
                mqueue_name = cls.ALL
        else:
            access = cls.ALL
            mqueue_type = cls.ALL
            label = cls.ALL
            mqueue_name = cls.ALL

        return cls(access, mqueue_type, label, mqueue_name,
                   audit=audit, deny=deny, allow_keyword=allow_keyword, comment=comment, priority=priority)

    def get_clean(self, depth=0):
        '''return rule (in clean/default formatting)'''

        space = '  ' * depth

        if self.all_access:
            access = ''
        elif len(self.access) == 1:
            access = ' %s' % ' '.join(self.access)
        elif self.access:
            access = ' (%s)' % ' '.join(sorted(self.access))
        else:
            raise AppArmorBug('Empty access in mqueue rule')

        if self.all_mqueue_types:
            mqueue_type = ''
        elif self.mqueue_type:
            mqueue_type = ' type=%s' % self.mqueue_type.regex
        else:
            raise AppArmorBug('Empty type in mqueue rule')

        if self.all_labels:
            label = ''
        elif self.label:
            label = ' label=%s' % quote_if_needed(self.label.regex)
        else:
            raise AppArmorBug('Empty label in mqueue rule')

        if self.all_mqueue_names:
            mqueue_name = ''
        elif self.mqueue_name:
            mqueue_name = ' %s' % self.mqueue_name.regex
        else:
            raise AppArmorBug('Empty mqueue_name in mqueue rule')

        return '%s%smqueue%s%s%s%s,%s' % (space, self.modifiers_str(), access, mqueue_type, label, mqueue_name, self.comment)

    def _is_covered_localvars(self, other_rule):
        '''check if other_rule is covered by this rule object'''

        if not self._is_covered_list(self.access, self.all_access, other_rule.access, other_rule.all_access, 'access'):
            return False

        if not self._is_covered_aare(self.mqueue_type, self.all_mqueue_types, other_rule.mqueue_type, other_rule.all_mqueue_types, 'mqueue_type'):
            return False

        if not self._is_covered_aare(self.label, self.all_labels, other_rule.label, other_rule.all_labels, 'label'):
            return False

        if not self._is_covered_aare(self.mqueue_name, self.all_mqueue_names, other_rule.mqueue_name, other_rule.all_mqueue_names, 'mqueue_name'):
            return False

        # still here? -> then it is covered
        return True

    def _is_equal_localvars(self, other_rule, strict):
        '''compare if rule-specific variables are equal'''

        if (self.access != other_rule.access or self.all_access != other_rule.all_access):
            return False

        if not self._is_equal_aare(self.mqueue_type, self.all_mqueue_types, other_rule.mqueue_type, other_rule.all_mqueue_types, 'mqueue_type'):
            return False

        if not self._is_equal_aare(self.label, self.all_labels, other_rule.label, other_rule.all_labels, 'label'):
            return False

        if not self._is_equal_aare(self.mqueue_name, self.all_mqueue_names, other_rule.mqueue_name, other_rule.all_mqueue_names, 'mqueue_name'):
            return False

        return True

    def _logprof_header_localvars(self):
        access = logprof_value_or_all(self.access, self.all_access)
        mqueue_type = logprof_value_or_all(self.mqueue_type, self.all_mqueue_types)
        label = logprof_value_or_all(self.label, self.all_labels)
        mqueue_name = logprof_value_or_all(self.mqueue_name, self.all_mqueue_names)

        return (
            _('Access mode'), access,
            _('Type'), mqueue_type,
            _('Label'), label,
            _('Message queue name'), mqueue_name
        )

    @staticmethod
    def hashlog_from_event(hl, ev):
        mqueue_type = ev['class'].partition('_')[0]
        hl[ev['denied_mask']][mqueue_type][ev['name']] = True

    @classmethod
    def from_hashlog(cls, hl):
        for access, mqueue_type, mqueue_name in BaseRule.generate_rules_from_hashlog(hl, 3):
            yield cls(access, mqueue_type, MessageQueueRule.ALL, mqueue_name, log_event=True)


class MessageQueueRuleset(BaseRuleset):
    '''Class to handle and store a collection of mqueue rules'''

    def get_glob(self, path_or_rule):
        '''Return the next possible glob. For mqueue rules, that means removing access, label or mqueue_name'''
        # XXX only remove one part, not all
        return 'mqueue,'



# ----------------------------------------------------------------------
#    Copyright (C) 2013 Kshitij Gupta <kgupta8592@gmail.com>
#    Copyright (C) 2020 Christian Boltz <apparmor@cboltz.de>
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------

from apparmor.common import AppArmorBug, AppArmorException
from apparmor.regex import RE_PROFILE_BOOLEAN
from apparmor.rule import BaseRule, BaseRuleset, parse_comment
from apparmor.translations import init_translation

_ = init_translation()


class BooleanRule(BaseRule):
    """Class to handle and store a single variable rule"""

    rule_name = 'boolean'
    _match_re = RE_PROFILE_BOOLEAN

    def __init__(self, varname, value, audit=False, deny=False, allow_keyword=False,
                 comment='', log_event=None, priority=None):

        super().__init__(audit=audit, deny=deny, allow_keyword=allow_keyword,
                         comment=comment, log_event=log_event, priority=priority)

        # boolean variables don't support priority, allow keyword, audit or deny
        self.ensure_modifiers_not_supported()

        if not isinstance(varname, str):
            raise AppArmorBug('Passed unknown type for boolean variable to %s: %s' % (self.__class__.__name__, varname))
        if not varname.startswith('$'):
            raise AppArmorException("Passed invalid boolean to %s (doesn't start with '$'): %s" % (self.__class__.__name__, varname))

        if not isinstance(value, str):
            raise AppArmorBug('Passed unknown type for value to %s: %s' % (self.__class__.__name__, value))
        if not value:
            raise AppArmorException('Passed empty value to %s: %s' % (self.__class__.__name__, value))

        value = value.lower()
        if value not in ('true', 'false'):
            raise AppArmorException('Passed invalid value to %s: %s' % (self.__class__.__name__, value))

        self.varname = varname
        self.value = value == 'true'

    @classmethod
    def _create_instance(cls, raw_rule, matches):
        """parse raw_rule and return instance of this class"""

        comment = parse_comment(matches)

        varname = matches.group('varname')
        value = matches.group('value')

        return cls(varname, value,
                   audit=False, deny=False, allow_keyword=False, comment=comment, priority=None)

    def get_clean(self, depth=0):
        """return rule (in clean/default formatting)"""

        space = '  ' * depth

        return '%s%s = %s' % (space, self.varname, str(self.value).lower())

    def _is_covered_localvars(self, other_rule):
        """check if other_rule is covered by this rule object"""

        if self.varname != other_rule.varname:
            return False

        if type(self.value) is not bool or type(other_rule.value) is not bool:
            raise AppArmorBug(_('Invalid value for boolean variable'))

        if self.value != other_rule.value:
            return False

        # still here? -> then it is covered
        return True

    def _is_equal_localvars(self, other_rule, strict):
        """compare if rule-specific variables are equal"""

        if self.varname != other_rule.varname:
            return False

        if self.value != other_rule.value:
            return False

        return True

    def _logprof_header_localvars(self):
        return _('Boolean Variable'), self.get_clean()


class BooleanRuleset(BaseRuleset):
    """Class to handle and store a collection of variable rules"""

    def add(self, rule, cleanup=False):
        """Add boolean variable rule object

           If the variable name is already known, raise an exception because re-defining a variable isn't allowed.
        """

        for knownrule in self.rules:
            if rule.varname == knownrule.varname:
                raise AppArmorException(_('Redefining existing variable %(variable)s: %(value)s')
                                        % {'variable': rule.varname, 'value': rule.value})

        super().add(rule, cleanup)



# ----------------------------------------------------------------------
#    Copyright (C) 2022 Canonical, Ltd.
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------

import re

from apparmor.regex import RE_PROFILE_USERNS
from apparmor.common import AppArmorBug, AppArmorException
from apparmor.rule import BaseRule, BaseRuleset, check_and_split_list, logprof_value_or_all, parse_modifiers

# setup module translations
from apparmor.translations import init_translation
_ = init_translation()

access_keyword = 'create'

RE_USERNS_DETAILS = re.compile(
    '^'
    + r'\s+(?P<access>' + access_keyword + ')?'  # optional access keyword
    + r'\s*$')


class UserNamespaceRule(BaseRule):
    '''Class to handle and store a single userns rule'''

    # Nothing external should reference this class, all external users
    # should reference the class field UserNamespaceRule.ALL
    class __UserNamespaceAll(object):
        pass

    ALL = __UserNamespaceAll

    rule_name = 'userns'
    _match_re = RE_PROFILE_USERNS

    def __init__(self, access, audit=False, deny=False,
                 allow_keyword=False, comment='', log_event=None, priority=None):

        super().__init__(audit=audit, deny=deny,
                         allow_keyword=allow_keyword,
                         comment=comment,
                         log_event=log_event,
                         priority=priority)

        self.access, self.all_access, unknown_items = check_and_split_list(access, access_keyword, self.ALL, type(self).__name__, 'access')
        if unknown_items:
            raise AppArmorException(_('Passed unknown access keyword to %s: %s') % (type(self).__name__, ' '.join(unknown_items)))

    @classmethod
    def _create_instance(cls, raw_rule, matches):
        '''parse raw_rule and return instance of this class'''

        priority, audit, deny, allow_keyword, comment = parse_modifiers(matches)

        rule_details = ''
        if matches.group('details'):
            rule_details = matches.group('details')

        if rule_details:
            details = RE_USERNS_DETAILS.search(rule_details)
            if not details:
                raise AppArmorException(_("Invalid or unknown keywords in 'userns %s" % rule_details))

            access = details.group('access')
        else:
            access = cls.ALL

        return cls(access, audit=audit, deny=deny,
                   allow_keyword=allow_keyword, comment=comment, priority=priority)

    @staticmethod
    def hashlog_from_event(hl, ev):
        if ev['denied_mask']:
            hl[ev['denied_mask'][7:]] = True  # [7:] removes the 'userns_' prefix
        else:
            hl[ev['request_mask'][7:]] = True  # To support transition to special profiles

    @classmethod
    def from_hashlog(cls, hl):
        for access in BaseRule.generate_rules_from_hashlog(hl, 1):
            yield cls(access)

    def get_clean(self, depth=0):
        '''return rule (in clean/default formatting)'''

        space = '  ' * depth

        if self.all_access:
            access = ''
        elif self.access:
            access = ' %s' % ' '.join(self.access)
        else:
            raise AppArmorBug('Empty access in userns rule')

        return '%s%suserns%s,%s' % (space, self.modifiers_str(), access, self.comment)

    def _is_covered_localvars(self, other_rule):
        '''check if other_rule is covered by this rule object'''

        if not self._is_covered_list(self.access, self.all_access, other_rule.access, other_rule.all_access, 'access'):
            return False

        # still here? -> then it is covered
        return True

    def _is_equal_localvars(self, other_rule, strict):
        '''compare if rule-specific variables are equal'''

        if (self.access != other_rule.access or self.all_access != other_rule.all_access):
            return False

        return True

    def _logprof_header_localvars(self):
        access = logprof_value_or_all(self.access, self.all_access)

        return _('Access mode'), access


class UserNamespaceRuleset(BaseRuleset):
    '''Class to handle and store a collection of userns rules'''

    def get_glob(self, path_or_rule):
        '''Return the next possible glob. For userns rules, that means removing access'''
        return 'userns,'



# ----------------------------------------------------------------------
#    Copyright (C) 2015 Christian Boltz <apparmor@cboltz.de>
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------

import re

from apparmor.common import AppArmorBug, AppArmorException
from apparmor.regex import RE_PROFILE_PTRACE, RE_PROFILE_NAME, strip_quotes
from apparmor.rule import (
    BaseRule, BaseRuleset, check_and_split_list, logprof_value_or_all,
    parse_modifiers, quote_if_needed)
from apparmor.translations import init_translation

_ = init_translation()

access_keywords = ['r', 'w', 'rw', 'wr', 'read', 'write', 'readby', 'trace', 'tracedby']  # XXX 'wr' and 'write' accepted by the parser, but not documented in apparmor.d.pod

# XXX joint_access_keyword and RE_ACCESS_KEYWORDS exactly as in PtraceRule - move to function!
joint_access_keyword = r'\s*(' + '|'.join(access_keywords) + r')\s*'
RE_ACCESS_KEYWORDS = (joint_access_keyword  # one of the access_keyword
                      + '|'  # or
                      + r'\(' + joint_access_keyword + '(' + r'(\s|,)+' + joint_access_keyword + ')*' + r'\)')  # one or more access_keyword in (...)


RE_PTRACE_DETAILS = re.compile(
    '^'
    + r'(\s+(?P<access>' + RE_ACCESS_KEYWORDS + '))?'  # optional access keyword(s)
    + r'(\s+(peer=' + RE_PROFILE_NAME % 'peer' + '))?'  # optional peer
    + r'\s*$')


class PtraceRule(BaseRule):
    """Class to handle and store a single ptrace rule"""

    # Nothing external should reference this class, all external users
    # should reference the class field PtraceRule.ALL
    class __PtraceAll:
        pass

    ALL = __PtraceAll

    rule_name = 'ptrace'
    _match_re = RE_PROFILE_PTRACE

    def __init__(self, access, peer, audit=False, deny=False, allow_keyword=False,
                 comment='', log_event=None, priority=None):

        super().__init__(audit=audit, deny=deny, allow_keyword=allow_keyword,
                         comment=comment, log_event=log_event, priority=priority)

        self.access, self.all_access, unknown_items = check_and_split_list(
            access, access_keywords, self.ALL, type(self).__name__, 'access')
        if unknown_items:
            raise AppArmorException(_('Passed unknown access keyword to %s: %s') % (type(self).__name__, ' '.join(unknown_items)))

        self.peer, self.all_peers = self._aare_or_all(peer, 'peer', is_path=False, log_event=log_event)

    @classmethod
    def _create_instance(cls, raw_rule, matches):
        """parse raw_rule and return instance of this class"""

        priority, audit, deny, allow_keyword, comment = parse_modifiers(matches)

        rule_details = ''
        if matches.group('details'):
            rule_details = matches.group('details')

        if rule_details:
            details = RE_PTRACE_DETAILS.search(rule_details)
            if not details:
                raise AppArmorException(_("Invalid or unknown keywords in 'ptrace %s" % rule_details))

            if details.group('access'):
                # XXX move to function _split_access()?
                access = details.group('access')
                if access.startswith('(') and access.endswith(')'):
                    access = access[1:-1]
                access = access.replace(',', ' ').split()  # split by ',' or whitespace
            else:
                access = cls.ALL

            if details.group('peer'):
                peer = strip_quotes(details.group('peer'))
            else:
                peer = cls.ALL
        else:
            access = cls.ALL
            peer = cls.ALL

        return cls(access, peer,
                   audit=audit, deny=deny, allow_keyword=allow_keyword, comment=comment, priority=priority)

    def get_clean(self, depth=0):
        """return rule (in clean/default formatting)"""

        space = '  ' * depth

        if self.all_access:
            access = ''
        elif len(self.access) == 1:
            access = ' %s' % ' '.join(self.access)
        elif self.access:
            access = ' (%s)' % ' '.join(sorted(self.access))
        else:
            raise AppArmorBug('Empty access in ptrace rule')

        if self.all_peers:
            peer = ''
        elif self.peer:
            peer = ' peer=%s' % quote_if_needed(self.peer.regex)
        else:
            raise AppArmorBug('Empty peer in ptrace rule')

        return ('%s%sptrace%s%s,%s' % (space, self.modifiers_str(), access, peer, self.comment))

    def _is_covered_localvars(self, other_rule):
        """check if other_rule is covered by this rule object"""

        if not self._is_covered_list(self.access, self.all_access, other_rule.access, other_rule.all_access, 'access'):
            return False

        if not self._is_covered_aare(self.peer, self.all_peers, other_rule.peer, other_rule.all_peers, 'peer'):
            return False

        # still here? -> then it is covered
        return True

    def _is_equal_localvars(self, other_rule, strict):
        """compare if rule-specific variables are equal"""

        if (self.access != other_rule.access
                or self.all_access != other_rule.all_access):
            return False

        if not self._is_equal_aare(self.peer, self.all_peers, other_rule.peer, other_rule.all_peers, 'peer'):
            return False

        return True

    def _logprof_header_localvars(self):
        access = logprof_value_or_all(self.access, self.all_access)
        peer   = logprof_value_or_all(self.peer,   self.all_peers)  # noqa: E221

        return (
            _('Access mode'), access,
            _('Peer'), peer,
        )

    @staticmethod
    def hashlog_from_event(hl, ev):
        hl[ev['peer']][ev['denied_mask']] = True

    @classmethod
    def from_hashlog(cls, hl):
        for peer in hl.keys():
            if '//null-' in peer:
                continue  # ignore null-* peers

            for access in hl[peer].keys():
                yield cls(access, peer, log_event=True)


class PtraceRuleset(BaseRuleset):
    """Class to handle and store a collection of ptrace rules"""

    def get_glob(self, path_or_rule):
        """Return the next possible glob. For ptrace rules, that means removing access or removing/globbing peer"""
        # XXX only remove one part, not all
        return 'ptrace,'



# ----------------------------------------------------------------------
#    Copyright (C) 2013 Kshitij Gupta <kgupta8592@gmail.com>
#    Copyright (C) 2014-2024 Christian Boltz <apparmor@cboltz.de>
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------

from types import NoneType

from apparmor.common import AppArmorBug, AppArmorException
from apparmor.regex import parse_profile_start_line, re_print_dict
from apparmor.rule import quote_if_needed
from apparmor.rule.abi import AbiRule, AbiRuleset
from apparmor.rule.all import AllRule, AllRuleset
from apparmor.rule.capability import CapabilityRule, CapabilityRuleset
from apparmor.rule.change_profile import ChangeProfileRule, ChangeProfileRuleset
from apparmor.rule.dbus import DbusRule, DbusRuleset
from apparmor.rule.file import FileRule, FileRuleset
from apparmor.rule.include import IncludeRule, IncludeRuleset
from apparmor.rule.network import NetworkRule, NetworkRuleset
from apparmor.rule.ptrace import PtraceRule, PtraceRuleset
from apparmor.rule.rlimit import RlimitRule, RlimitRuleset
from apparmor.rule.signal import SignalRule, SignalRuleset
from apparmor.rule.userns import UserNamespaceRule, UserNamespaceRuleset
from apparmor.rule.mqueue import MessageQueueRule, MessageQueueRuleset
from apparmor.rule.io_uring import IOUringRule, IOUringRuleset
from apparmor.rule.mount import MountRule, MountRuleset
from apparmor.rule.pivot_root import PivotRootRule, PivotRootRuleset
from apparmor.rule.unix import UnixRule, UnixRuleset
from apparmor.rule.conditional import ConditionalBlock, ConditionalBlockset

from apparmor.translations import init_translation

_ = init_translation()

ruletypes = {
    'abi':            {'rule': AbiRule,           'ruleset': AbiRuleset},
    'inc_ie':         {'rule': IncludeRule,       'ruleset': IncludeRuleset},
    'all':            {'rule': AllRule,           'ruleset': AllRuleset},
    'capability':     {'rule': CapabilityRule,    'ruleset': CapabilityRuleset},
    'change_profile': {'rule': ChangeProfileRule, 'ruleset': ChangeProfileRuleset},
    'dbus':           {'rule': DbusRule,          'ruleset': DbusRuleset},
    'file':           {'rule': FileRule,          'ruleset': FileRuleset},
    'network':        {'rule': NetworkRule,       'ruleset': NetworkRuleset},
    'ptrace':         {'rule': PtraceRule,        'ruleset': PtraceRuleset},
    'rlimit':         {'rule': RlimitRule,        'ruleset': RlimitRuleset},
    'signal':         {'rule': SignalRule,        'ruleset': SignalRuleset},
    'userns':         {'rule': UserNamespaceRule, 'ruleset': UserNamespaceRuleset},
    'mqueue':         {'rule': MessageQueueRule,  'ruleset': MessageQueueRuleset},
    'io_uring':       {'rule': IOUringRule,       'ruleset': IOUringRuleset},
    'mount':          {'rule': MountRule,         'ruleset': MountRuleset},
    'pivot_root':     {'rule': PivotRootRule,     'ruleset': PivotRootRuleset},
    'unix':           {'rule': UnixRule,          'ruleset': UnixRuleset},
    'cond_block':     {'rule': ConditionalBlock,  'ruleset': ConditionalBlockset},
}


class ProfileStorage:
    """class to store the content (header, rules, comments) of a profilename

       Acts like a dict(), but has some additional checks.
    """

    def __init__(self, profilename, hat, calledby):
        data = dict()

        data['info'] = {'profile': profilename, 'hat': hat, 'calledby': calledby}

        for rule in ruletypes:
            data[rule] = ruletypes[rule]['ruleset']()

        data['filename'] = ''
        data['logprof_suggest'] = ''  # set in abstractions that should be suggested by aa-logprof
        data['parent'] = ''  # parent profile, or '' for top-level profiles and external hats
        data['name'] = ''
        data['attachment'] = ''
        data['xattrs'] = {}
        data['flags'] = ''
        data['external'] = False
        data['header_comment'] = ''  # comment in the profile/hat start line
        data['initial_comment'] = ''
        data['profile_keyword'] = False
        data['is_hat'] = False  # profile or hat?
        data['hat_keyword'] = False  # True for 'hat foo', False for '^foo'
        data['in_cond'] = False  # used for profiles created inside conditionals

        self.data = data

    def __getitem__(self, key):
        if key in self.data:
            return self.data[key]
        else:
            raise AppArmorBug('attempt to read unknown key %s' % key)

    def __setitem__(self, key, value):
        if key not in self.data:
            raise AppArmorBug('attempt to set unknown key %s' % key)

        allowed_types = {bool, str, dict, None, NoneType}
        old_type = type(self.data[key])
        if old_type in allowed_types:
            if key in {'flags', 'filename'} and type(value) in {str, NoneType}:
                self.data[key] = value
            elif isinstance(value, old_type):
                self.data[key] = value
            else:
                raise AppArmorBug('Attempt to change type of "%s" from %s to %s, value %s' % (key, old_type, type(value), value))
            self.data[key] = value
        else:
            raise AppArmorBug('Attempt to overwrite "%s" with %s, type %s' % (key, value, type(value)))

    def __repr__(self):
        classname = type(self).__name__
        header = '\n'.join(self.get_header(0, False))
        rules = '\n'.join(self.get_rules_clean(1))
        endprofile = '}'
        return f'\n<{classname}>\n{header}\n{rules}\n{endprofile}\n</{classname}>\n'

    def get(self, key, fallback=None):
        if key in self.data:
            return self.data.get(key, fallback)
        else:
            raise AppArmorBug('attempt to read unknown key %s' % key)

    def get_header(self, depth, embedded_hat):
        pre = ' ' * int(depth * 2)
        data = []
        unquoted_name = self.data['info']['profile']
        name = quote_if_needed(self.data['info']['profile'])

        attachment = ''
        if self.data['attachment']:
            attachment = ' %s' % quote_if_needed(self.data['attachment'])

        comment = ''
        if self.data['header_comment']:
            comment = ' %s' % self.data['header_comment']

        if self.data['external']:
            name = quote_if_needed(unquoted_name + '//' + self.data['info']['hat'])

        if not self.data['external'] and self.data['parent']:  # child profile
            name = quote_if_needed(self.data['info']['hat'])

        if self.data['is_hat']:
            if self.data['hat_keyword']:
                name = 'hat %s' % name
            else:
                name = '^%s' % name
        elif (not embedded_hat and not unquoted_name.startswith('/')) or (embedded_hat and not unquoted_name.startswith('^')) or self.data['attachment'] or self.data['profile_keyword']:
            name = 'profile %s%s' % (name, attachment)

        xattrs = ''
        if self.data['xattrs']:
            xattrs = ' xattrs=(%s)' % re_print_dict(self.data['xattrs'])

        flags = ''
        if self.data['flags']:
            flags = ' flags=(%s)' % self.data['flags']

        data.append('%s%s%s%s {%s' % (pre, name, xattrs, flags, comment))

        return data

    def get_rules_clean(self, depth):
        """return all clean rules of a profile (with default formatting, and leading whitespace as specified in the depth parameter)

           Note that the profile header and the closing "}" are _not_ included.
        """

        write_order = [
            'abi',
            'inc_ie',
            'rlimit',
            'capability',
            'cond_block',
            'network',
            'dbus',
            'mount',
            'mqueue',
            'signal',
            'ptrace',
            'pivot_root',
            'unix',
            'file',
            'change_profile',
            'userns',
            'io_uring',
        ]

        data = []

        for ruletype in write_order:
            data.extend(self.data[ruletype].get_clean(depth))

        return data

    def get_local_include(self):
        inc = None
        preferred_inc = self.data['name']
        if preferred_inc.startswith('/'):
            preferred_inc = preferred_inc[1:]
        preferred_inc = 'local/' + preferred_inc.replace('/', '.')

        # If a local profile already exists, we use it.
        for rule in self.data['inc_ie'].rules:
            if rule.path.startswith("local/"):
                inc = rule.path
                if rule.path == preferred_inc:  # Prefer includes that matches the profile name.
                    break
        return inc

    @classmethod
    def parse(cls, line, file, lineno, profile, hat):
        """parse a profile start line (using parse_profile_startline()) and convert it to an instance of this class"""

        matches = parse_profile_start_line(line, file)

        if profile:  # we are inside a profile, so we expect a child profile
            if not matches['profile_keyword']:
                raise AppArmorException(
                    _('%(profile)s profile in %(file)s contains syntax errors in line %(line)s: missing "profile" keyword.')
                    % {'profile': profile, 'file': file, 'line': lineno + 1})
            if hat is not None:
                # nesting limit reached - a child profile can't contain another child profile
                raise AppArmorException(
                    _('%(profile)s profile in %(file)s contains syntax errors in line %(line)s: a child profile inside another child profile is not allowed.')
                    % {'profile': profile, 'file': file, 'line': lineno + 1})

            parent = profile
            hat = matches['profile']
            prof_or_hat_name = hat
            pps_set_hat_external = False

        else:  # stand-alone profile
            parent = ''
            profile = matches['profile']
            prof_or_hat_name = profile
            if len(profile.split('//')) > 2:
                raise AppArmorException(
                    "Nested child profiles ('%(profile)s', found in %(file)s) are not supported by the AppArmor tools yet."
                    % {'profile': profile, 'file': file})
            elif len(profile.split('//')) == 2:
                profile, hat = profile.split('//')
                parent = profile
                pps_set_hat_external = True
            else:
                hat = profile
                pps_set_hat_external = False

        prof_storage = cls(profile, hat, cls.__name__ + '.parse()')

        prof_storage['parent'] = parent
        prof_storage['name'] = prof_or_hat_name
        prof_storage['filename'] = file
        prof_storage['external'] = pps_set_hat_external
        prof_storage['flags'] = matches['flags']
        prof_storage['header_comment'] = matches['comment'] or ''
        prof_storage['is_hat'] = matches['is_hat']

        if matches['is_hat']:
            prof_storage['hat_keyword'] = matches['hat_keyword']
        else:
            prof_storage['profile_keyword'] = matches['profile_keyword']
            prof_storage['attachment'] = matches['attachment'] or ''
            prof_storage['xattrs'] = matches['xattrs'] or {}

        return (profile, hat, prof_storage)


def split_flags(flags):
    """split the flags given as string into a sorted, de-duplicated list"""

    if flags is None:
        flags = ''

    # Flags may be whitespace and/or comma separated
    flags_list = flags.replace(',', ' ').split()
    # sort and remove duplicates
    return sorted(set(flags_list))


def add_or_remove_flag(flags, flags_to_change, set_flag):
    """add (if set_flag is True) or remove the given flags_to_change to flags"""

    if isinstance(flags, str) or flags is None:
        flags = split_flags(flags)

    if isinstance(flags_to_change, str) or flags_to_change is None:
        flags_to_change = split_flags(flags_to_change)

    if set_flag:
        for flag_to_change in flags_to_change:
            if flag_to_change not in flags:
                flags.append(flag_to_change)
    else:
        for flag_to_change in flags_to_change:
            if flag_to_change in flags:
                flags.remove(flag_to_change)

    return sorted(flags)


def var_transform(ref):
    data = []
    for value in sorted(ref):
        if not value:
            value = '""'
        data.append(quote_if_needed(value))
    return ' '.join(data)
